@@ -1,6 +1,6 @@
 /***********************************************************************
  *
- * Copyright (C) 2009, 2010, 2011, 2012, 2013, 2014, 2015 Graeme Gott <graeme@gottcode.org>
+ * Copyright (C) 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016, 2017, 2018, 2019 Graeme Gott <graeme@gottcode.org>
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -32,6 +32,8 @@
 #include <QCloseEvent>
 #include <QDialog>
 #include <QDialogButtonBox>
+#include <QDragEnterEvent>
+#include <QDropEvent>
 #include <QFile>
 #include <QFileDialog>
 #include <QFormLayout>
@@ -40,10 +42,12 @@
 #include <QMenu>
 #include <QMenuBar>
 #include <QMessageBox>
+#include <QMimeData>
 #include <QSettings>
 #include <QStackedWidget>
 #include <QStandardPaths>
 #include <QStyle>
+#include <QTemporaryFile>
 #include <QTextEdit>
 #include <QTextStream>
 #include <QVBoxLayout>
@@ -274,8 +278,10 @@ namespace
 
 //-----------------------------------------------------------------------------
 
-Window::Window()
+Window::Window(const QString& file)
 : m_pause_action(0), m_previous_state(0) {
+	setAcceptDrops(true);
+
 	// Create states
 	m_states.insert("NewGame", new NewGameState(this));
 	m_states.insert("OpenGame", new OpenGameState(this));
@@ -398,13 +404,21 @@ Window::Window()
 	counts_action->setChecked(settings.value("ShowWordCounts", true).toBool());
 	restoreGeometry(settings.value("Geometry").toByteArray());
 
-	// Start a new game
+	// Start game
+	QString current = file;
+	if (settings.contains("Current/Version")) {
+		if (current.isEmpty() ||
+				QMessageBox::question(this, tr("Question"), tr("End the current game?"), QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::No) {
+			current = ":saved:";
+		}
+	}
+
 	m_state->finish();
 	m_contents->setCurrentIndex(3);
-	if (!settings.contains("Current/Version")) {
+	if (current.isEmpty()) {
 		newGame();
 	} else {
-		startGame(":saved:");
+		startGame(current);
 	}
 }
 
@@ -439,6 +453,75 @@ void Window::closeEvent(QCloseEvent* event) {
 
 //-----------------------------------------------------------------------------
 
+void Window::dragEnterEvent(QDragEnterEvent* event) {
+	if (!(event->possibleActions() & Qt::CopyAction)) {
+		return;
+	}
+
+	// Handle dragged file
+	if (event->mimeData()->hasUrls()) {
+		const QList<QUrl> urls = event->mimeData()->urls();
+		if (urls.size() != 1) {
+			return;
+		}
+		const QUrl url = urls.first();
+		if (!url.isLocalFile()) {
+			return;
+		}
+		const QString filename = url.toLocalFile();
+		if (!filename.endsWith(".tanglet")) {
+			return;
+		}
+	// Handle dragged game data
+	} else if (!event->mimeData()->hasFormat("application/x-tanglet")) {
+		return;
+	}
+
+	event->setDropAction(Qt::CopyAction);
+	event->acceptProposedAction();
+}
+
+//-----------------------------------------------------------------------------
+
+void Window::dropEvent(QDropEvent* event) {
+	if (!(event->possibleActions() & Qt::CopyAction)) {
+		return;
+	}
+
+	// Handle dropped file
+	if (event->mimeData()->hasUrls()) {
+		const QList<QUrl> urls = event->mimeData()->urls();
+		if (urls.size() != 1) {
+			return;
+		}
+		const QUrl url = urls.first();
+		if (!url.isLocalFile()) {
+			return;
+		}
+		const QString filename = url.toLocalFile();
+		if (!filename.endsWith(".tanglet")) {
+			return;
+		}
+		if (endGame()) {
+			startGame(filename);
+		}
+	// Handle dropped game data
+	} else if (event->mimeData()->hasFormat("application/x-tanglet")) {
+		QTemporaryFile file;
+		if (file.open()) {
+			file.write(event->mimeData()->data("application/x-tanglet"));
+		}
+		if (endGame()) {
+			startGame(file.fileName());
+		}
+	}
+
+	event->setDropAction(Qt::CopyAction);
+	event->acceptProposedAction();
+}
+
+//-----------------------------------------------------------------------------
+
 bool Window::event(QEvent* event) {
 	if (m_pause_action && m_pause_action->isEnabled()) {
 		switch (event->type()) {
@@ -467,7 +550,7 @@ void Window::about() {
 		QString("<center><p><big><b>%1</b></big><br/>%2<br/><small>%3<br/>%4</small></p><p>%5</p><p>%6</p></center>")
 		.arg(tr("Tanglet %1").arg(QCoreApplication::applicationVersion()),
 			tr("A single player variant of <a href=\"http://en.wikipedia.org/wiki/Boggle\">Boggle</a>"),
-			tr("Copyright &copy; 2009-%1 Graeme Gott").arg("2015"),
+			tr("Copyright &copy; 2009-%1 Graeme Gott").arg("2019"),
 			tr("Released under the <a href=\"http://www.gnu.org/licenses/gpl.html\">GPL 3</a> license"),
 			tr("English word list is based on <a href=\"http://wordlist.sourceforge.net/\">SCOWL</a> by Kevin Atkinson"),
 			tr("Hebrew word list is based on <a href=\"http://hspell.ivrix.org.il/\">Hspell</a> by Nadav Har'El and Dan Kenigsberg"))
@@ -517,54 +600,7 @@ void Window::chooseGame() {
 				QStandardPaths::writableLocation(QStandardPaths::DocumentsLocation),
 				tr("Tanglet Games (*.tanglet)"));
 		if (!filename.isEmpty()) {
-			try
-			{
-				QString current = QStandardPaths::writableLocation(QStandardPaths::DataLocation);
-
-				// Uncompress shared game
-				{
-					QDir dir = QDir::home();
-					dir.mkpath(current);
-					current += "/current";
-
-					QByteArray data = gunzip(filename);
-					QFile file(current);
-					if (!file.open(QFile::WriteOnly)) {
-						throw tr("Unable to start requested game.");
-					}
-					file.write(data);
-					file.close();
-				}
-
-				// Extract words and dice
-				{
-					QSettings game(current, QSettings::IniFormat);
-
-					if (game.value("Game/Language", -1).toInt() == 0) {
-						QByteArray dice = QByteArray::fromBase64(game.value("Game/Dice").toByteArray());
-						game.setValue("Game/Dice", current + "-dice");
-						QFile file(current + "-dice");
-						if (!file.open(QFile::WriteOnly)) {
-							throw tr("Unable to start requested game.");
-						}
-						file.write(dice);
-						file.close();
-
-						QByteArray words = QByteArray::fromBase64(game.value("Game/Words").toByteArray());
-						game.setValue("Game/Words", current + "-words");
-						file.setFileName(current + "-words");
-						if (!file.open(QFile::WriteOnly)) {
-							throw tr("Unable to start requested game.");
-						}
-						file.write(words);
-						file.close();
-					}
-				}
-
-				startGame(current);
-			} catch (const QString& error) {
-				QMessageBox::warning(this, tr("Error"), error);
-			}
+			startGame(filename);
 		}
 	}
 }
@@ -773,9 +809,15 @@ void Window::startGame(const QString& filename) {
 			game = new QSettings;
 			game->beginGroup("Current");
 		} else {
+			QString file = extractGame(filename);
+			if (file.isEmpty()) {
+				QMessageBox::warning(this, tr("Error"), tr("Unable to start requested game."));
+				return;
+			}
+
 			// Start requested game
 			QSettings().remove("Current");
-			game = new QSettings(filename, QSettings::IniFormat);
+			game = new QSettings(file, QSettings::IniFormat);
 			game->beginGroup("Game");
 		}
 
@@ -796,3 +838,47 @@ void Window::startGame(const QString& filename) {
 }
 
 //-----------------------------------------------------------------------------
+
+QString Window::extractGame(const QString& filename) const {
+	QString current = QStandardPaths::writableLocation(QStandardPaths::DataLocation);
+
+	// Uncompress shared game
+	const QDir dir = QDir::home();
+	dir.mkpath(current);
+	current += "/current";
+
+	const QByteArray data = gunzip(filename);
+	QFile file(current);
+	if (!file.open(QFile::WriteOnly)) {
+		return QString();
+	}
+	file.write(data);
+	file.close();
+
+	// Extract words and dice
+	QSettings game(current, QSettings::IniFormat);
+
+	if (game.value("Game/Language", -1).toInt() == 0) {
+		const QByteArray dice = QByteArray::fromBase64(game.value("Game/Dice").toByteArray());
+		game.setValue("Game/Dice", current + "-dice");
+		file.setFileName(current + "-dice");
+		if (!file.open(QFile::WriteOnly)) {
+			return QString();
+		}
+		file.write(dice);
+		file.close();
+
+		const QByteArray words = QByteArray::fromBase64(game.value("Game/Words").toByteArray());
+		game.setValue("Game/Words", current + "-words");
+		file.setFileName(current + "-words");
+		if (!file.open(QFile::WriteOnly)) {
+			return QString();
+		}
+		file.write(words);
+		file.close();
+	}
+
+	return current;
+}
+
+//-----------------------------------------------------------------------------
